@@ -22,9 +22,156 @@ This is an HTML generator.
 package HTML::Make;
 use warnings;
 use strict;
-our $VERSION = 0.01;
+our $VERSION = 0.02;
 use Carp;
 
+# Extracted from
+
+# https://developer.mozilla.org/en-US/docs/Web/HTML/Element?redirectlocale=en-US&redirectslug=HTML%2FElement
+
+# by "get-tags.pl".
+
+my %tags = qw/
+a 1
+abbr 1
+acronym 1
+address 1
+applet 1
+area 1
+article 1
+aside 1
+audio 1
+b 1
+base 1
+basefont 1
+bdi 1
+bdo 1
+bgsound 1
+big 1
+blink 1
+blockquote 1
+body 1
+br 1
+button 1
+canvas 1
+caption 1
+center 1
+cite 1
+code 1
+col 1
+colgroup 1
+content 1
+data 1
+datalist 1
+dd 1
+decorator 1
+del 1
+details 1
+dfn 1
+dir 1
+div 1
+dl 1
+dt 1
+element 1
+em 1
+embed 1
+fieldset 1
+figcaption 1
+figure 1
+font 1
+footer 1
+form 1
+frame 1
+frameset 1
+h1 1
+h2 1
+h3 1
+h4 1
+h5 1
+h6 1
+head 1
+header 1
+hgroup 1
+hr 1
+html 1
+i 1
+iframe 1
+img 1
+input 1
+ins 1
+isindex 1
+kbd 1
+keygen 1
+label 1
+legend 1
+li 1
+link 1
+listing 1
+main 1
+map 1
+mark 1
+marquee 1
+menu 1
+menuitem 1
+meta 1
+meter 1
+nav 1
+nobr 1
+noframes 1
+noscript 1
+object 1
+ol 1
+optgroup 1
+option 1
+output 1
+p 1
+param 1
+plaintext 1
+pre 1
+progress 1
+q 1
+rp 1
+rt 1
+ruby 1
+s 1
+samp 1
+script 1
+section 1
+select 1
+shadow 1
+small 1
+source 1
+spacer 1
+span 1
+strike 1
+strong 1
+style 1
+sub 1
+summary 1
+sup 1
+table 1
+tbody 1
+td 1
+template 1
+textarea 1
+tfoot 1
+th 1
+thead 1
+time 1
+title 1
+tr 1
+track 1
+tt 1
+u 1
+ul 1
+var 1
+video 1
+wbr 1
+xmp 1
+/;
+
+our $texttype = 'text';
+
 =head1 METHODS
 
 =cut
@@ -35,14 +182,25 @@ use Carp;
 
 Make a new HTML element of the specified type.
 
-It is possible to add attributes or text to the item.
+It is possible to add attributes or text to the item. To add
+attributes, use the following syntax:
 
     my $element = HTML::Make->new ('li', attr => {class => 'biglist'});
 
-Add text:
+To add text,
 
     my $element = HTML::Make->new ('li', text => "White punks on dope");
 
+Both attributes and text may be added:
+
+    my $element = HTML::Make->new ('li', attr => {id => 'ok'}, text => 'OK');
+
+HTML::Make has a list of known HTML tags and will issue a warning if
+the type given as the first argument to new is not on its list of
+tags. To switch off this behaviour, use
+
+    my $freaky = HTML::Make->new ('freaky', nocheck => 1);
+
 =cut
 
 sub new
@@ -51,17 +209,30 @@ sub new
     my $obj = {};
     bless $obj;
     $obj->{type} = $type;
-    if ($options{text}) {
-        if ($type eq 'text') {
-            $obj->{text} = $options{text};
-        }
-        else {
+    # User is not allowed to use 'text' type.
+    if ($type eq $texttype) {
+	my ($package, undef, undef) = caller ();
+	if ($package ne __PACKAGE__) {
+	    die "Illegal use of text type";
+	}
+	if (! defined $options{text}) {
+	    die "Text type object with empty text";
+	}
+	if (ref $options{text}) {
+	    croak "text field must be a scalar";
+	}
+	$obj->{text} = $options{text};
+    }
+    else {
+	if (! $options{nocheck} && ! $tags{lc $type}) {
+	    carp "Unknown tag type '$type'";
+	}
+	if ($options{text}) {
             $obj->add_text ($options{text});
         }
-    }
-    
-    if ($options{attr}) {
-        $obj->add_attr (%{$options{attr}});
+	if ($options{attr}) {
+	    $obj->add_attr (%{$options{attr}});
+	}
     }
     return $obj;
 }
@@ -77,12 +248,19 @@ Add attributes to the specified object, in other words
     my $obj_text = $obj->text ();
     # <li class="beano"></li>
 
+This issues a warning of the form B<"Overwriting attribute 'class' for
+'li'"> if the object already contains an attribute of the specified
+type.
+
 =cut
 
 sub add_attr
 {
     my ($obj, %attr) = @_;
     for my $k (keys %attr) {
+	if ($obj->{attr}->{$k}) {
+	    carp "Overwriting attribute '$k' for '$obj->{type}' tag";
+	}
         $obj->{attr}->{$k} = $attr{$k};
     }
 }
@@ -110,8 +288,8 @@ The text may contain HTML elements:
 sub add_text
 {
     my ($obj, $text) = @_;
-    my $x = __PACKAGE__->new ('text', text => $text);
-    push @{$obj->{children}}, $x;
+    my $x = __PACKAGE__->new ($texttype, text => $text);
+    CORE::push @{$obj->{children}}, $x;
     return $x;
 }
 
@@ -128,7 +306,8 @@ as a new C<HTML::Make> object. For example,
     print $table->text ();
     # <table><tr><td></td></tr></table>
 
-It's also possible to add the same arguments as L</new>:
+It's also possible to add all of the same arguments as L</new>, for
+example
 
     $element->push ('a', attr => {href => 'http://www.example.org/'});
 
@@ -138,7 +317,7 @@ sub HTML::Make::push
 {
     my ($obj, $el, %options) = @_;
     my $x = __PACKAGE__->new ($el, %options);
-    push @{$obj->{children}}, $x;
+    CORE::push @{$obj->{children}}, $x;
     return $x;
 }
 
@@ -154,20 +333,19 @@ attributes.
 sub opening_tag
 {
     my ($obj) = @_;
-    my $text = '';
-        $text .= "<$obj->{type}";
-        if ($obj->{attr}) {
-            my @attr;
-            my %attr = %{$obj->{attr}};
-            for my $k (keys %attr) {
-                my $v = $attr{$k};
-                $v =~ s/"/\\"/g;
-                CORE::push @attr, "$k=\"$v\"";
-            }
-            my $attr = join (' ', @attr);
-            $text .= " $attr";
-        }
-        $text .= ">";
+    my $text = "<$obj->{type}";
+    if ($obj->{attr}) {
+	my @attr;
+	my %attr = %{$obj->{attr}};
+	for my $k (keys %attr) {
+	    my $v = $attr{$k};
+	    $v =~ s/"/\\"/g;
+	    CORE::push @attr, "$k=\"$v\"";
+	}
+	my $attr = join (' ', @attr);
+	$text .= " $attr";
+    }
+    $text .= ">";
     return $text;
 }
 
@@ -190,16 +368,16 @@ sub text
     if (! $type) {
         croak "No type";
     }
-    my $text = '';
-    if ($type eq 'text') {
+    my $text;
+    if ($type eq $texttype) {
         $text = $obj->{text};
     }
     else {
-        $text .= $obj->opening_tag ();
+        $text = $obj->opening_tag ();
         for my $child (@{$obj->{children}}) {
             $text .= $child->text ();
         }
-        $text .= "</$obj->{type}>\n";
+        $text .= "</$type>\n";
     }
     return $text;
 }
@@ -232,6 +410,9 @@ sub multiply
         my $x = $parent->push ($element, text => $content);
         CORE::push @elements, $x;
     }
+    if (@elements != @$contents) {
+	die "Mismatch of number of elements";
+    }
     return @elements;
 }
 
@@ -243,4 +424,4 @@ Copyright (c) 2013 Ben Bullock. This Perl module may be used,
 redistributed, modified, and copied under the same terms as Perl
 itself.
 
-
+=cut
